from ..exception import StructHasDefinedError, StructNotFoundError


class Registry:
    def __init__(self):
        self.struct_map = {}

    def register(self, struct_name, struct_cls):
        # NOTICE: Since this method is called when models are imported,
        # it cannot perform imports because of the risk of import loops.
        if struct_name in self.struct_map:
            raise StructHasDefinedError
        self.struct_map[struct_name] = struct_cls

    def get_struct(self, name):
        if name not in self.struct_map:
            raise StructNotFoundError
        return self.struct_map[name]


registry = Registry()



import click

class OptionEatAll(click.Option):
    """
    implemented by https://stackoverflow.com/users/7311767/stephen-rauch in https://stackoverflow.com/questions/48391777/nargs-equivalent-for-options-in-click
    """

    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop('save_other_options', True)
        nargs = kwargs.pop('nargs', -1)
        assert nargs == -1, 'nargs, if set, must be -1 not {}'.format(nargs)
        super(OptionEatAll, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):

        def parser_process(value, state):
            # method to hook to the parser.process
            done = False
            value = [value]
            if self.save_other_options:
                # grab everything up to the next option
                while state.rargs and not done:
                    for prefix in self._eat_all_parser.prefixes:
                        if state.rargs[0].startswith(prefix):
                            done = True
                    if not done:
                        value.append(state.rargs.pop(0))
            else:
                # grab everything remaining
                value += state.rargs
                state.rargs[:] = []
            value = tuple(value)

            # call the actual process
            self._previous_parser_process(value, state)

        retval = super(OptionEatAll, self).add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break
        return retval


from .box import BBox
from .label import Label, LabelList
from .media import ImageMedia
from .polygon import Polygon, PolygonItem

__all__ = [
    "BBox",
    "Label",
    "ImageMedia",
    "LabelList",
    "Polygon",
    "PolygonItem",
]



import click
from abc import ABC, abstractmethod
from yaml import load as yaml_load
from .exception import DefineSyntaxError
from .warning import DuplicateDefineWarning
import networkx as nx
import os
import re


try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader


class Parser(ABC):
    """
    定义一个Parser的抽象基类（没法实例化的）：
    用处： 强制子类实现某些方法：写框架时常用。
    """

    @abstractmethod
    def parse(self, input_file_list):
        """
        将yaml文件中的模型（struct）和标签(label)部分校验之后读入某个变量中
        """
        pass

    @abstractmethod
    def generate(self, output_file):
        """
        将内存里面的模型（struct）和标签(label)部分输出成ORM模型（python代码）
        """
        pass

    def process(self, data_file, library_path, output_file):
        self.parse(data_file, library_path)
        self.generate(output_file)
        print(
            f"Convert Yaml File to Python Code Successfully!\n \
            Yaml file (source): {data_file}\n \
            Output file (output): {output_file}"
        )


class DSDLParser(Parser):
    def __init__(self):
        self.define_map = dict()
        self.TYPES_WITHOUT_PARS = [
            "Bool",
            "Num",
            "Int",
            "Str",
            "Coord",
            "Coord3D",
            "Interval",
            "BBox",
            "Polygon",
            "Image",
        ]
        self.TYPES_WITH_PARS = ["Date", "Label", "Time"]
        self.TYPES_WITH_PARS_SP = ["List"]
        self.dsdl_version = None
        self.data_sample_type = None
        self.sample_type_par_map = dict()

    def parse(self, data_file, library_path):
        """
        将yaml文件中的模型（struct）和标签(label)部分校验之后读入变量self.define_map中
            input_file_list: 读入的yaml文件
        """
        with open(data_file, "r") as f:
            desc = yaml_load(f, Loader=YAMLSafeLoader)
        # 获取 self.data_sample_type和self.sample_type_par_map
        try:
            data_sample_type = desc["data"]["sample-type"]
        except KeyError as e:
            raise DefineSyntaxError(f"data yaml must contains {e} in `data` section")
        self.parse_param_sample_type(data_sample_type)

        if "$import" in desc:
            import_list = desc["$import"]
            import_list = [
                os.path.join(library_path, p.strip() + ".yaml") for p in import_list
            ]
        else:
            import_list = []
        if "defs" in desc:
            # 获取yaml中模型（struct）和标签(label)部分的内容，存储在变量class_defi中，
            # 因为有不同格式的yaml(数据和模型放同一个yaml中或者分开放)，所以用if...else分别做处理
            # 注意区分这个root_class_defi,为啥要把他先存好？如果import的yaml中，有重复的模型，需要用它覆盖，参见白皮书2.5.1
            root_class_defi = desc["defs"].items()
            self.dsdl_version = desc["$dsdl-version"]  # 存版本号，后续应该会使用（目前木有用）
        else:
            root_class_defi = dict()

        import_desc = dict()
        for input_file in import_list:
            with open(input_file, "r") as f:
                import_desc.update(yaml_load(f, Loader=YAMLSafeLoader))

        if "defs" in import_desc:
            # 获取yaml中模型（struct）和标签(label)部分的内容，存储在变量class_defi中，
            # 因为有不同格式的yaml(数据和模型放同一个yaml中或者分开放)，所以用if...else分别做处理
            class_defi = import_desc["defs"]
        else:
            class_defi = import_desc
        # root_class_defi是数据yaml里面定义的模型，如果和import里面的重复了，会覆盖掉前面import的。参见白皮书2.5.1
        class_defi.update(root_class_defi)

        # 对class_defi循环，处理里面的每一个struct或者label(class_domain)
        for define_name, define_value in class_defi.items():
            if define_name == "$dsdl-version":
                self.dsdl_version = define_value  # 存版本号，后续应该会使用（目前木有用）
                continue
            define_type = define_value["$def"]
            if not define_name.isidentifier():
                continue
            if define_name in self.define_map:
                DuplicateDefineWarning(f"{define_name} has defined.")

            define_info = {"name": define_name}
            if define_type == "struct":
                # 对class_defi中struct类型的ele做校验并存入define_info
                define_info["type"] = "struct"
                if "$params" in define_value:
                    replace_flag = True
                else:
                    replace_flag = False
                define_info["field_list"] = []
                for raw_field in define_value["$fields"].items():
                    if not raw_field[0].isidentifier():
                        continue
                    field = raw_field[1].replace(" ", "")
                    if replace_flag:
                        for param, value in self.sample_type_par_map.items():
                            field = field.replace("$" + param, value)
                    define_info["field_list"].append(
                        {
                            "name": raw_field[0],
                            "type": self.parse_struct_field(field),
                        }
                    )

            if define_type == "class_domain":
                # 对class_defi中class_domain类型的ele（也就是定义的label）做校验并存入define_info
                define_info["type"] = "class_domain"
                define_info["class_list"] = []
                for class_name in define_value["classes"]:
                    define_info["class_list"].append(
                        {
                            "name": self.clean(class_name),
                        }
                    )

            self.define_map[define_info["name"]] = define_info

    def generate(self, output_file):
        """
        将内存里面的模型（struct）和标签(label)部分输出成ORM模型（python代码）
        """
        # check define cycles. 如果有环形（就是循环定义）那是不行滴～
        define_graph = nx.DiGraph()
        define_graph.add_nodes_from(self.define_map.keys())
        for key, val in self.define_map.items():
            if val["type"] != "struct":
                continue
            for field in val["field_list"]:
                for k in self.define_map.keys():
                    if k in field["type"]:
                        define_graph.add_edge(k, key)
        if not nx.is_directed_acyclic_graph(define_graph):
            raise "define cycle found."

        with open(output_file, "w") as of:
            print("# Generated by the dsdl parser. DO NOT EDIT!\n", file=of)
            print("from dsdl.types import *\nfrom enum import Enum, unique\n\n", file=of)
            ordered_keys = list(nx.topological_sort(define_graph))
            for idx, key in enumerate(ordered_keys):
                val = self.define_map[key]
                if val["type"] == "struct":
                    print(f"class {key}(Struct):", file=of)
                    for field in val["field_list"]:
                        print(f"""    {field["name"]} = {field["type"]}""", file=of)
                if val["type"] == "class_domain":
                    print("@unique", file=of)
                    print(f"class {key}(Enum):", file=of)
                    for class_id, item in enumerate(val["class_list"], start=1):
                        class_name = item["name"]
                        print(f'''    {class_name.lower()} = {class_id}''', file=of)
                if idx != len(ordered_keys) - 1:
                    print("\n", file=of)

    def parse_param_sample_type(self, raw: str):
        """
        raw: ObjectDetectionSample[cdom=COCOClassFullDom]
        """
        c_dom = re.findall(r"\[(.*?)\]", raw)
        if c_dom:
            temp = c_dom[0].split("=")
            self.sample_type_par_map[temp[0].strip()] = temp[1].strip()
            self.data_sample_type = raw.replace("[" + c_dom[0] + "]", "", 1).strip()
        else:
            self.data_sample_type = raw

    def parse_list_filed(self, raw: str) -> str:
        """
        解析处理List类型的field
        """

        def sanitize_etype(val: str) -> str:
            """
            验证List类型中的etype是否存在（必须存在）且是否为合法类型
            """
            return self.parse_struct_field(val)

        def all_subclasses(cls):
            """
            返回某个类的所有子类：like[<class '__main__.Bar'>, <class '__main__.Baz'>...]
            """
            return cls.__subclasses__() + [
                g for s in cls.__subclasses__() for g in all_subclasses(s)
            ]

        def rreplace(s, old, new, occurrence):
            """
            从右向左的替换函数，类似replace,不过是反着的
            """
            li = s.rsplit(old, occurrence)
            return new.join(li)

        def sanitize_ordered(val: str) -> str:
            if val.lower() in ["true", "false"]:
                if val.lower() == "true":
                    return "True"
                else:
                    return "False"
            else:
                raise DefineSyntaxError(
                    f"invalid value {val} in ordered of List {raw}."
                )

        field_type = "List"

        raw = rreplace(raw.replace(f"{field_type}[", "", 1), "]", "", 1)
        param_list = raw.split(",")
        ele_type, ordered = None, None
        if len(param_list) == 2:
            ele_type = param_list[0]
            ordered = param_list[1]
        elif len(param_list) == 1:
            ele_type = param_list[0]
        else:
            raise DefineSyntaxError(f"invalid parameters {raw} in List.")

        ele_type = ele_type.split("=", 1)
        if len(ele_type) == 2:
            if ele_type[0].strip() != "etype":
                raise DefineSyntaxError(f"List types must contains parameters `etype`.")
            ele_type = ele_type[1]
            c_dom = re.findall(r"\[(.*?)\]", ele_type)
            if c_dom:
                ele_type = ele_type.replace("[" + c_dom[0] + "]", "", 1).strip()
        else:
            ele_type = ele_type[0]
        # else:
        #     raise DefineSyntaxError(f"invalid parameters {raw} in List.")

        res = field_type + "Field("
        if ele_type:
            ele_type = sanitize_etype(ele_type)
            res += "ele_type=" + ele_type
        else:
            raise DefineSyntaxError(f"List types must contains parameters `etype`.")
        if ordered:
            ordered = ordered.split("=")[-1]
            ordered = sanitize_ordered(ordered)
            res += ", ordered=" + ordered
        return res + ")"

    @staticmethod
    def parse_struct_field_with_params(raw: str) -> str:
        """
        解析处理Label, Time, Date类型的field
        """

        def sanitize_dom(val: str) -> str:
            """
            Label中对dom部分的校验，是用来严格限制特定格式或者字符。
            防止yaml里有一些异常代码注入到生成的Python 代码里被执行起来。
            """
            if not val.isidentifier():
                raise DefineSyntaxError(f"invalid dom: {val}")
            return val

        def sanitize_fmt(val: str) -> str:
            """
            Date, Time中对ftm部分的校验，是用来严格限制特定格式或者字符。
            防止yaml里有一些异常代码注入到生成的Python 代码里被执行起来。
            """
            val = val.strip("\"'")
            return f'"{val}"'

        field_map = {
            "Label": {"dom": sanitize_dom},
            "Date": {"fmt": sanitize_fmt},
            "Time": {"fmt": sanitize_fmt},
        }
        field_type = ""
        for k in field_map.keys():
            if raw.startswith(k):
                field_type = k
        if field_type == "":
            raise "Unknown field"

        raw = raw.replace(f"{field_type}[", "").replace("]", "")
        param_list = raw.split(",")
        valid_param_list = []
        for param in param_list:
            parts = param.split("=")
            # 需要考虑参数省略的情况，因为dom经常省略
            if len(parts) == 2:
                field_para = parts[0]
                field_var = parts[1]
            elif len(parts) == 1:
                field_para = next(iter(field_map[field_type]))
                field_var = parts[0]
            else:
                raise DefineSyntaxError(f"invalid parameters {raw} in List.")
            sanitized = field_map[field_type][field_para](field_var)
            valid_param_list.append(f"{field_para}={sanitized}")
        return field_type + "Field(" + ",".join(valid_param_list) + ")"

    def parse_struct_field(self, raw_field_type: str) -> str:
        """
        校验struct类型的每个字段的入口函数，对不同情况（Int,Image,List...）的字段进行校验并读入内存。
            raw_field_type: like: Int, Image, Label[dom=MyClassDom], List[List[Int], ordered = True], ....
        """
        if raw_field_type in self.TYPES_WITHOUT_PARS:
            # 不带参数的不需要校验，直接可以转化为python代码（注意区分yaml中的模型部分到ORM的校验和通过ORM读入数据的校验，
            # 后者在dataset.py中定义，你脑子里想的Int也需要校验是数据的校验。）
            return raw_field_type + "Field()"
        elif raw_field_type.startswith(tuple(self.TYPES_WITH_PARS_SP)):
            # 带参数的Date, Time, Label类型的字段的校验
            return self.parse_list_filed(raw_field_type)
        elif raw_field_type.startswith(tuple(self.TYPES_WITH_PARS)):
            # 带参数的List类型的字段的校验
            # （因为List类型比较特殊，里面可以包含List，Label等各种其他字段，涉及递归，所以单独拿出来）
            return DSDLParser.parse_struct_field_with_params(raw_field_type)
        else:
            return raw_field_type + "()"
            # raise DefineTypeError(f"No type {raw_field_type} in DSDL.")

    @staticmethod
    def clean(varStr):
        return re.sub('\W|^(?=\d)', '_', varStr)


@click.command()
@click.option(
    "-y",
    "--yaml",
    "dsdl_yaml",
    type=str,
    required=True,
)
@click.option(
    "-p",
    "--path",
    "dsdl_library_path",
    type=str,
    default="dsdl/dsdl_library",
)
def parse(dsdl_yaml, dsdl_library_path):
    dsdl_name = os.path.splitext(os.path.basename(dsdl_yaml))[0]
    output_file = os.path.join(os.path.dirname(dsdl_yaml), f"{dsdl_name}.py")
    dsdl_parser = DSDLParser()
    dsdl_parser.process(dsdl_yaml, dsdl_library_path, output_file)



class Field:
    def __init__(self):
        pass

    def __str__(self):
        return "<%s>" % self.__class__.__name__

    def validate(self, value):
        """
        Validate value and raise ValidationError if necessary.
        """
        return value



from dsdl.tools import DetectionParse, Generation
from argparse import ArgumentParser
import os


def parse_args():
    parser = ArgumentParser()
    parser.add_argument("-i", "--dataset_info", type=str, required=True,
                        help="the path of dataset_info.json of tunas 0.3 format dataset.")
    parser.add_argument("-a", "--ann_info", type=str, required=True,
                        help="the path of annotation.json of tunas 0.3 format dataset.")
    parser.add_argument("-w", "--working_dir", type=str, required=True,
                        help="the working dir you want to save the result file.")
    return parser.parse_args()


def main(dataset_info, annotation_info, working_dir):
    conversion = DetectionParse(dataset_info, annotation_info)
    print(conversion.struct_defs)
    print(conversion.class_domain_info)
    print(conversion.meta_info)
    print(conversion.dsdl_version)
    generate_obj = Generation(conversion.dsdl_version, conversion.meta_info, conversion.struct_defs,
                              conversion.class_domain_info, conversion.samples, working_dir)
    class_file = generate_obj.write_class_dom()
    def_file = generate_obj.write_struct_def(file_name="object-detection")
    generate_obj.write_samples(file_name=os.path.basename(annotation_info).replace(".json", ""),
                               import_list=[class_file, def_file])


if __name__ == '__main__':
    args = parse_args()
    main(args.dataset_info, args.ann_info, args.working_dir)



import re
from .field import Field
from ..geometry import BBox, Label, Polygon, PolygonItem
from ..exception import ValidationError
from ..warning import InvalidLabelWarning
from datetime import date, time, datetime


def validate_list_of_number(value, size_limit, item_type):
    if type(value) is not list:
        raise ValidationError(f"expect list of num, got {value}")
    if len(value) != size_limit:
        raise ValidationError(f"expect size of list is {size_limit}, got {len(value)}")
    try:
        return [item_type(item) for item in value]
    except (TypeError, ValueError) as _:
        raise ValidationError(f"expect type of list item is float, got {value}")


class CoordField(Field):
    def validate(self, value):
        return validate_list_of_number(value, 2, float)


class Coord3DField(Field):
    def validate(self, value):
        return validate_list_of_number(value, 3, float)


class IntervalField(Field):
    def validate(self, value):
        value = validate_list_of_number(value, 2, float)
        if value[0] > value[1]:
            raise ValidationError(
                f"expect |begin| less than or equal to |end|, got {value}"
            )
        return value


class BBoxField(Field):
    def validate(self, value):
        return BBox(*validate_list_of_number(value, 4, float))


class PolygonField(Field):
    def validate(self, value):
        polygon_lst = []
        for idx, points in enumerate(value):
            for point in points:
                validate_list_of_number(point, 2, float)
            polygon_lst.append(PolygonItem(points))
        return Polygon(polygon_lst)


class LabelField(Field):
    def __init__(self, dom):
        super(LabelField, self).__init__()
        self.dom = dom

    def validate(self, value):
        try:
            if isinstance(value, int):
                category_name = self.dom(value).name
                return Label(category_name=category_name, category_id=value, class_domain=self.dom)
            elif isinstance(value, str):
                category_id = getattr(self.dom, self.clean(value)).value
                return Label(category_name=value, category_id=category_id, class_domain=self.dom)
            else:
                raise TypeError("invalid class label type.")
        except:
            InvalidLabelWarning(f"The label {value} is not valid.")
            return Label(category_name="invalid", category_id=-1, class_domain=self.dom)

    @staticmethod
    def clean(varStr):
        return re.sub('\W|^(?=\d)', '_', varStr).lower()



class DateField(Field):
    def __init__(self, fmt: str = ""):
        super(DateField, self).__init__()
        self.fmt = fmt

    def validate(self, value):
        if self.fmt == "":
            return date.fromisoformat(value)
        return datetime.strptime(value, self.fmt).date()


class TimeField(Field):
    def __init__(self, fmt: str = ""):
        super(TimeField, self).__init__()
        self.fmt = fmt

    def validate(self, value):
        if self.fmt == "":
            return time.fromisoformat(value)
        return datetime.strptime(value, self.fmt).time()


class AttributesField(Field):
    def __init__(self, content: Field):
        super(AttributesField).__init__()
        self._content_type = content

    def validate(self, value):
        return self._content_type.validate(value)



import io
from PIL import Image
import numpy as np


class ImageMedia:

    def __init__(self, location, file_reader):
        self._loc = location
        self._reader = file_reader

    @property
    def location(self):
        return self._loc

    def to_bytes(self):
        return io.BytesIO(self._reader.read())

    def to_image(self):
        return Image.open(self.to_bytes())

    def to_array(self):
        return np.array(self.to_image())

    def __repr__(self):
        return f"path:{self.location} size:{self.to_image().size}"



import numpy as np
from PIL import ImageFont, ImageDraw


class Label:

    def __init__(self, category_id, category_name, class_domain):

        self._id = category_id
        self._name = category_name
        self._dom = class_domain

    @property
    def category_name(self):
        return self._name

    @property
    def category_id(self):
        return self._id

    @property
    def openmmlabformat(self):
        return self._name

    @property
    def class_domain(self):
        return self._dom

    def visualize(self, image, palette, **kwargs):
        draw_obj = ImageDraw.Draw(image)
        if self.category_name not in palette:
            palette[self.category_name] = tuple(np.random.randint(0, 255, size=[3]))
        color = palette[self.category_name]
        font = ImageFont.load_default()
        label_size = draw_obj.textsize(self.category_name, font)
        if "bbox" in kwargs:
            coords = np.array([[item.xyxy[0], item.xyxy[1] + 0.2 * label_size[1]] for item in kwargs["bbox"].values()])
        elif "polygon" in kwargs:
            coords = np.array([[item.point_for_draw[0], item.point_for_draw[1] + 0.2 * label_size[1]] for item in
                               kwargs["polygon"].values()])
        else:
            coords = np.array([[0, 0.2 * label_size[1]]])
        for coord in coords:
            draw_obj.rectangle([tuple(coord), tuple(coord + label_size)], fill=(*color, 255))
            draw_obj.text(tuple(coord), self.category_name, fill=(255, 255, 255, 255), font=font)
        del draw_obj
        return image

    def __repr__(self):
        return self.category_name


class LabelList:

    def __init__(self, label_list):
        self._label_list = list(label_list)

    @property
    def category_names(self):
        return [label.category_name for label in self._label_list]

    @property
    def category_ids(self):
        return [label.category_id for label in self._label_list]

    @property
    def label_list(self):
        return self._label_list

    @property
    def class_domains(self):
        return [label.class_domain for label in self._label_list]

    def visualize(self, image, palette, **kwargs):
        draw_obj = ImageDraw.Draw(image)
        font = ImageFont.load_default()
        y_offset = np.zeros((1, 2))
        for label_obj in self.label_list:

            category_name = label_obj.category_name

            if category_name not in palette:
                palette[category_name] = tuple(np.random.randint(0, 255, size=[3]))
            color = palette[category_name]

            label_size = draw_obj.textsize(category_name, font)
            if "bbox" in kwargs:
                # coords.shape = [num_box, 2]
                coords = y_offset + np.array(
                    [[item.xyxy[0], item.xyxy[1] + 0.2 * label_size[1]] for item in kwargs["bbox"].values()])
            elif "polygon" in kwargs:
                coords = y_offset + np.array(
                    [[item.point_for_draw[0], item.point_for_draw[1] + 0.2 * label_size[1]] for item in
                     kwargs["polygon"].values()])
            elif "image_label_list" not in kwargs:
                coords = y_offset + np.array([[0, 0.2 * label_size[1]]])
            else:
                coords = []
                kwargs["image_label_list"].append(label_obj)

            y_offset += np.array([[0., 1.2 * label_size[1]]])
            for coord in coords:
                draw_obj.rectangle([tuple(coord), tuple(coord + label_size)], fill=(*color, 255))
                draw_obj.text(tuple(coord), category_name, fill=(255, 255, 255, 255), font=font)

        del draw_obj
        return image



from .field import Field
from .registry import registry
from ..exception import ValidationError


class StructMetaclass(type):
    def __new__(mcs, name, bases, attributes):
        super_new = super().__new__

        # Also ensure initialization is only performed for subclasses of Model
        # (excluding Model class itself).
        parents = [b for b in bases if isinstance(b, StructMetaclass)]
        if not parents:
            return super_new(mcs, name, bases, attributes)

        mappings = dict()
        for k, v in attributes.items():
            if isinstance(v, Field):
                mappings[k] = v
        for k in mappings.keys():
            attributes.pop(k)
        attributes["__mappings__"] = mappings
        new_class = super_new(mcs, name, bases, attributes)
        registry.register(name, new_class)
        return new_class


class Struct(dict, metaclass=StructMetaclass):
    def __init__(self, file_reader=None, **kwargs):
        super().__init__()
        self.file_reader = file_reader
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(r"'Model' object has no attribute '%s'" % key)

    def __setattr__(self, key, value):

        if key not in self.__mappings__:
            self[key] = value
            return

        if hasattr(self.__mappings__[key], "set_file_reader"):
            self.__mappings__[key].set_file_reader(self.file_reader)
        try:
            self[key] = self.__mappings__[key].validate(value)
        except ValidationError as error:
            raise ValidationError(f"Field '{key}' validation error: {error}.")

    def get_mapping(self):
        return self.__mappings__



import os

try:
    import ruamel_yaml as yaml
    from ruamel_yaml import YAML
except ImportError:
    import ruamel.yaml as yaml
    from ruamel.yaml import YAML


class Generation:
    LINESEP_NUM = 1

    def __init__(self, dsdl_version, meta_info, def_info, class_info, data_info, working_dir):
        self._working_dir = working_dir
        self._dsdl_version = dsdl_version
        self._meta_info = meta_info
        self._def_info = def_info
        self._data_info = data_info
        self._class_info = class_info
        self._yaml_obj = self._prepare_yaml_obj()

    def _prepare_yaml_obj(self):
        default_yaml = YAML()
        class_dom_yaml = YAML()
        class_dom_yaml.indent(mapping=4, sequence=4, offset=4)
        sample_yaml = YAML()
        sample_yaml.indent(mapping=4, sequence=4, offset=2)

        result = {"class_dom": class_dom_yaml, "dsdl_version": default_yaml, "meta_info": default_yaml,
                  "struct_def": class_dom_yaml, "samples": sample_yaml, "import": class_dom_yaml}
        return result

    def write_meta(self, file_name=None, path=None):
        meta_info = self._meta_info.copy()
        for k, v in meta_info.items():
            meta_info[k] = self.add_quotes(v)
        if file_name is not None:
            path = os.path.join(self._working_dir, file_name + ".yaml")
        else:
            assert path is not None, "you must provide `file_name` or `path` args"
        with open(path, 'a') as fp:
            self._yaml_obj["meta_info"].dump(meta_info, fp)
            fp.writelines([os.linesep] * self.LINESEP_NUM)

    def write_dsdl_version(self, file_name=None, path=None):
        dsdl_version_info = self._dsdl_version.copy()
        for k, v in dsdl_version_info.items():
            dsdl_version_info[k] = self.add_quotes(v)
        if file_name is not None:
            path = os.path.join(self._working_dir, file_name + ".yaml")
        else:
            assert path is not None, "you must provide `file_name` or `path` args"
        with open(path, 'a') as fp:
            self._yaml_obj["dsdl_version"].dump(dsdl_version_info, fp)
            fp.writelines([os.linesep] * self.LINESEP_NUM)

    @staticmethod
    def add_quotes(string):
        # string = f'"{string}"'
        string = yaml.scalarstring.DoubleQuotedScalarString(string)
        return string

    @staticmethod
    def flist(x):
        x = [Generation.add_quotes(_) for _ in x]
        retval = yaml.comments.CommentedSeq(x)
        retval.fa.set_flow_style()  # fa -> format attribute
        return retval

    @staticmethod
    def fmap(x):
        retval = yaml.comments.CommentedMap(x)
        retval.fa.set_flow_style()  # fa -> format attribute
        return retval

    def write_class_dom(self, file_name=None):
        class_info = self._class_info.copy()
        class_dom_name = class_info.pop('$name')
        file_name = file_name if file_name else class_dom_name
        dst = os.path.join(self._working_dir, file_name + ".yaml")
        self.write_dsdl_version(path=dst)
        with open(dst, "a") as fp:
            self._yaml_obj["class_dom"].dump({class_dom_name: class_info}, fp)
            fp.writelines([os.linesep] * self.LINESEP_NUM)
        return file_name

    def _write_struct_item(self, struct_item, file_name=None, path=None):
        if file_name is not None:
            path = os.path.join(self._working_dir, file_name + ".yaml")
        else:
            assert path is not None, "you must provide `file_name` or `path` args"
        struct_item = struct_item.copy()
        if "$params" in struct_item and isinstance(struct_item["$params"], list):
            struct_item["$params"] = self.flist(struct_item["$params"])
        if "$optional" in struct_item and isinstance(struct_item["$optional"], list):
            struct_item["$optional"] = self.flist(struct_item["$optional"])
        struct_name = struct_item.pop("$name")
        with open(path, "a") as fp:
            self._yaml_obj["struct_def"].dump({struct_name: struct_item}, fp)
            fp.writelines([os.linesep] * self.LINESEP_NUM)

    @staticmethod
    def is_list_of(obj, cls):
        if isinstance(obj, list):
            return isinstance(obj[0], cls)
        return False

    def write_struct_def(self, file_name):

        dst = os.path.join(self._working_dir, file_name + ".yaml")
        self.write_dsdl_version(path=dst)

        for struct_def_item in self._def_info:
            self._write_struct_item(struct_def_item, file_name)
        return file_name

    def write_import_list(self, import_list=None, file_name=None, path=None):
        if not import_list:
            return
        if file_name is not None:
            path = os.path.join(self._working_dir, file_name + ".yaml")
        else:
            assert path is not None, "you must provide `file_name` or `path` args"
        with open(path, "a") as fp:
            self._yaml_obj["import"].dump({"$import": import_list}, fp)
            fp.writelines([os.linesep] * self.LINESEP_NUM)
        return file_name

    def write_samples(self, file_name, import_list=None):

        dst = os.path.join(self._working_dir, file_name + ".yaml")
        self.write_dsdl_version(path=dst)
        self.write_import_list(import_list, path=dst)
        data_info = {"sample-type": self._data_info["sample-type"], "samples": []}
        samples = self._data_info["samples"]
        for sample in samples:
            sample_ = {}
            for k, v in sample.items():
                if self.is_list_of(v, dict):
                    sample_[k] = [self.fmap(_) for _ in v]
                elif isinstance(v, str):
                    sample_[k] = self.add_quotes(v)
                else:
                    sample_[k] = v
            data_info["samples"].append(sample_)

        with open(dst, "a") as fp:
            self._yaml_obj["samples"].dump({"data": data_info}, fp)



from base_dataset import Dataset
from mmdet.datasets import DATASETS
from mmdet.datasets.pipelines import Compose
from dsdl.dataset import ImageVisualizePipeline
from mmcv.utils import print_log
from terminaltables import AsciiTable

from mmdet.core import eval_recalls
import numpy as np
from itertools import product
from functools import partial
from collections import OrderedDict

try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader


@DATASETS.register_module()
class DSDLDetectionDataset(Dataset):
    CLASSES = []

    PALETTE = []

    FIELD_REQUIRED = (
        "bbox",
        "label",
    )

    FIELD_ALLOWED = (
        "bbox",
        "label",
        "polygon",
    )

    FIELD_MAPPING = {
        "bbox": "bboxes",
        "label": "labels",
        "polygon": "masks",
    }

    FIELD_DEFAULT_VALUE = {
        "bbox": np.zeros((0, 4), dtype=np.float32),
        "label": np.array([], dtype=np.int64),
        "polygon": [],
    }

    FIELD_PROCESS_FUNC = {
        "bbox": partial(np.array, dtype=np.float32),
        "label": partial(np.array, dtype=np.int64),
        "polygon": lambda _: _,
    }

    def __init__(self,
                 ann_file,
                 pipeline,
                 test_mode=False,
                 filter_empty_gt=True,
                 location_config=None,
                 field_list=None,
                 ):
        self._init_placeholder()
        self.ann_file = ann_file
        self.test_mode = test_mode
        pipeline = Compose(pipeline)
        field_list = [_ for _ in field_list if _ in self.FIELD_ALLOWED]
        assert all([_ in field_list for _ in self.FIELD_REQUIRED]), f"`field_list` must contains {self.FIELD_REQUIRED}."
        self.field_list = field_list
        self.palette = {}
        super().__init__(ann_file, location_config, pipeline)
        self.data_infos = [s for sample_ in self.sample_list for s in sample_]
        for class_name, class_palette in self.palette.items():
            self.CLASSES.append(class_name)
            self.PALETTE.append(class_palette)

    def _init_placeholder(self):
        # set unused field in mmdet
        self.img_prefix = '',
        self.seg_prefix = None,
        self.proposals = None

    def _parse_struct(self, sample):
        sample = super()._parse_struct(sample)
        vis_sample = ImageVisualizePipeline(self.field_list, sample, self.palette).visualize_result
        assert len(
            vis_sample.keys()) >= 1, f"There are no images in the sample {sample}, which is not permitted."

        result_lst = []
        for vis_obj in vis_sample.values():

            image_obj = vis_obj.image
            if self.test_mode:
                result_lst.append({'filename': image_obj})
            else:
                field_list = self.field_list.copy()
                ground_truth_dic = {_: [] for _ in field_list}
                for ground_truth in vis_obj.ground_truths.values():
                    # {bbox: {b1: xxx, b2: xxx}, label: {l1: xxx, l2: xxx), polygon: {p1: xxx, p2: xxx}}
                    # 如果缺失必要标注类型，则跳过该项
                    if any([_ not in ground_truth for _ in self.FIELD_REQUIRED]):
                        continue
                    temp_dic = {_: [] for _ in field_list}
                    for field in field_list:
                        for field_item in ground_truth.get(field, {}).values():
                            temp_dic[field].append(field_item.openmmlabformat)
                    for gt_group in product(*[temp_dic[_] or [None] for _ in field_list]):
                        for i, gt in enumerate(gt_group):
                            ground_truth_dic[field_list[i]].append(gt)

                result = {
                    'filename': image_obj,
                    'ann': {
                        self.FIELD_MAPPING[_]: self.FIELD_PROCESS_FUNC[_](ground_truth_dic[_]) if ground_truth_dic[
                            _] else self.FIELD_DEFAULT_VALUE[_]
                        for _ in field_list
                    }
                }
                result_lst.append(result)
        return result_lst

    def get_ann_info(self, idx):
        """Get annotation by index.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Annotation info of specified index.
        """

        return self.data_infos[idx]['ann']

    def __getitem__(self, idx):
        """Get training/test data after pipeline.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Training/test data (with annotation if `test_mode` is set \
                True).
        """

        if self.test_mode:
            return self.prepare_test_img(idx)
        else:
            return self.prepare_train_img(idx)

    def pre_pipeline(self, results):
        """Prepare results dict for pipeline."""
        results['img_prefix'] = self.img_prefix
        results['seg_prefix'] = self.seg_prefix
        results['proposal_file'] = self.proposal_file
        results['bbox_fields'] = []
        results['mask_fields'] = []
        results['seg_fields'] = []

    def prepare_train_img(self, idx):
        """Get training data and annotations after pipeline.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Training data and annotation after pipeline with new keys \
                introduced by pipeline.
        """

        img_info = self.data_infos[idx]
        ann_info = self.get_ann_info(idx)
        results = dict(img_info=img_info, ann_info=ann_info)
        if self.proposals is not None:
            results['proposals'] = self.proposals[idx]
        self.pre_pipeline(results)
        return self.pipeline(results)

    def prepare_test_img(self, idx):
        """Get testing data after pipeline.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Testing data after pipeline with new keys introduced by \
                pipeline.
        """

        img_info = self.data_infos[idx]
        results = dict(img_info=img_info)
        if self.proposals is not None:
            results['proposals'] = self.proposals[idx]
        self.pre_pipeline(results)
        return self.pipeline(results)


    def evaluate_det_segm(self,
                          results,
                          result_files,
                          coco_gt,
                          metrics,
                          logger=None,
                          classwise=False,
                          proposal_nums=(100, 300, 1000),
                          iou_thrs=None,
                          metric_items=None):
        """Instance segmentation and object detection evaluation in COCO
        protocol.

        Args:
            results (list[list | tuple | dict]): Testing results of the
                dataset.
            result_files (dict[str, str]): a dict contains json file path.
            coco_gt (COCO): COCO API object with ground truth annotation.
            metric (str | list[str]): Metrics to be evaluated. Options are
                'bbox', 'segm', 'proposal', 'proposal_fast'.
            logger (logging.Logger | str | None): Logger used for printing
                related information during evaluation. Default: None.
            classwise (bool): Whether to evaluating the AP for each class.
            proposal_nums (Sequence[int]): Proposal number used for evaluating
                recalls, such as recall@100, recall@1000.
                Default: (100, 300, 1000).
            iou_thrs (Sequence[float], optional): IoU threshold used for
                evaluating recalls/mAPs. If set to a list, the average of all
                IoUs will also be computed. If not specified, [0.50, 0.55,
                0.60, 0.65, 0.70, 0.75, 0.80, 0.85, 0.90, 0.95] will be used.
                Default: None.
            metric_items (list[str] | str, optional): Metric items that will
                be returned. If not specified, ``['AR@100', 'AR@300',
                'AR@1000', 'AR_s@1000', 'AR_m@1000', 'AR_l@1000' ]`` will be
                used when ``metric=='proposal'``, ``['mAP', 'mAP_50', 'mAP_75',
                'mAP_s', 'mAP_m', 'mAP_l']`` will be used when
                ``metric=='bbox' or metric=='segm'``.

        Returns:
            dict[str, float]: COCO style evaluation metric.
        """
        if iou_thrs is None:
            iou_thrs = np.linspace(
                .5, 0.95, int(np.round((0.95 - .5) / .05)) + 1, endpoint=True)
        if metric_items is not None:
            if not isinstance(metric_items, list):
                metric_items = [metric_items]

        eval_results = OrderedDict()
        for metric in metrics:
            msg = f'Evaluating {metric}...'
            if logger is None:
                msg = '\n' + msg
            print_log(msg, logger=logger)

            if metric == 'proposal_fast':
                if isinstance(results[0], tuple):
                    raise KeyError('proposal_fast is not supported for '
                                   'instance segmentation result.')
                ar = self.fast_eval_recall(
                    results, proposal_nums, iou_thrs, logger='silent')
                log_msg = []
                for i, num in enumerate(proposal_nums):
                    eval_results[f'AR@{num}'] = ar[i]
                    log_msg.append(f'\nAR@{num}\t{ar[i]:.4f}')
                log_msg = ''.join(log_msg)
                print_log(log_msg, logger=logger)
                continue

            iou_type = 'bbox' if metric == 'proposal' else metric
            if metric not in result_files:
                raise KeyError(f'{metric} is not in results')
            try:
                predictions = mmcv.load(result_files[metric])
                if iou_type == 'segm':
                    # Refer to https://github.com/cocodataset/cocoapi/blob/master/PythonAPI/pycocotools/coco.py#L331  # noqa
                    # When evaluating mask AP, if the results contain bbox,
                    # cocoapi will use the box area instead of the mask area
                    # for calculating the instance area. Though the overall AP
                    # is not affected, this leads to different
                    # small/medium/large mask AP results.
                    for x in predictions:
                        x.pop('bbox')
                    warnings.simplefilter('once')
                    warnings.warn(
                        'The key "bbox" is deleted for more accurate mask AP '
                        'of small/medium/large instances since v2.12.0. This '
                        'does not change the overall mAP calculation.',
                        UserWarning)
                coco_det = coco_gt.loadRes(predictions)
            except IndexError:
                print_log(
                    'The testing results of the whole dataset is empty.',
                    logger=logger,
                    level=logging.ERROR)
                break

            cocoEval = COCOeval(coco_gt, coco_det, iou_type)
            cocoEval.params.catIds = self.cat_ids
            cocoEval.params.imgIds = self.img_ids
            cocoEval.params.maxDets = list(proposal_nums)
            cocoEval.params.iouThrs = iou_thrs
            # mapping of cocoEval.stats
            coco_metric_names = {
                'mAP': 0,
                'mAP_50': 1,
                'mAP_75': 2,
                'mAP_s': 3,
                'mAP_m': 4,
                'mAP_l': 5,
                'AR@100': 6,
                'AR@300': 7,
                'AR@1000': 8,
                'AR_s@1000': 9,
                'AR_m@1000': 10,
                'AR_l@1000': 11
            }
            if metric_items is not None:
                for metric_item in metric_items:
                    if metric_item not in coco_metric_names:
                        raise KeyError(
                            f'metric item {metric_item} is not supported')

            if metric == 'proposal':
                cocoEval.params.useCats = 0
                cocoEval.evaluate()
                cocoEval.accumulate()

                # Save coco summarize print information to logger
                redirect_string = io.StringIO()
                with contextlib.redirect_stdout(redirect_string):
                    cocoEval.summarize()
                print_log('\n' + redirect_string.getvalue(), logger=logger)

                if metric_items is None:
                    metric_items = [
                        'AR@100', 'AR@300', 'AR@1000', 'AR_s@1000',
                        'AR_m@1000', 'AR_l@1000'
                    ]

                for item in metric_items:
                    val = float(
                        f'{cocoEval.stats[coco_metric_names[item]]:.3f}')
                    eval_results[item] = val
            else:
                cocoEval.evaluate()
                cocoEval.accumulate()

                # Save coco summarize print information to logger
                redirect_string = io.StringIO()
                with contextlib.redirect_stdout(redirect_string):
                    cocoEval.summarize()
                print_log('\n' + redirect_string.getvalue(), logger=logger)

                if classwise:  # Compute per-category AP
                    # Compute per-category AP
                    # from https://github.com/facebookresearch/detectron2/
                    precisions = cocoEval.eval['precision']
                    # precision: (iou, recall, cls, area range, max dets)
                    assert len(self.cat_ids) == precisions.shape[2]

                    results_per_category = []
                    for idx, catId in enumerate(self.cat_ids):
                        # area range index 0: all area ranges
                        # max dets index -1: typically 100 per image
                        nm = self.coco.loadCats(catId)[0]
                        precision = precisions[:, :, idx, 0, -1]
                        precision = precision[precision > -1]
                        if precision.size:
                            ap = np.mean(precision)
                        else:
                            ap = float('nan')
                        results_per_category.append(
                            (f'{nm["name"]}', f'{float(ap):0.3f}'))

                    num_columns = min(6, len(results_per_category) * 2)
                    results_flatten = list(
                        itertools.chain(*results_per_category))
                    headers = ['category', 'AP'] * (num_columns // 2)
                    results_2d = itertools.zip_longest(*[
                        results_flatten[i::num_columns]
                        for i in range(num_columns)
                    ])
                    table_data = [headers]
                    table_data += [result for result in results_2d]
                    table = AsciiTable(table_data)
                    print_log('\n' + table.table, logger=logger)

                if metric_items is None:
                    metric_items = [
                        'mAP', 'mAP_50', 'mAP_75', 'mAP_s', 'mAP_m', 'mAP_l'
                    ]

                for metric_item in metric_items:
                    key = f'{metric}_{metric_item}'
                    val = float(
                        f'{cocoEval.stats[coco_metric_names[metric_item]]:.3f}'
                    )
                    eval_results[key] = val
                ap = cocoEval.stats[:6]
                eval_results[f'{metric}_mAP_copypaste'] = (
                    f'{ap[0]:.3f} {ap[1]:.3f} {ap[2]:.3f} {ap[3]:.3f} '
                    f'{ap[4]:.3f} {ap[5]:.3f}')

        return eval_results

    def evaluate(self,
                 results,
                 metric='bbox',
                 logger=None,
                 jsonfile_prefix=None,
                 classwise=False,
                 proposal_nums=(100, 300, 1000),
                 iou_thrs=None,
                 metric_items=None):
        """Evaluation in COCO protocol.

        Args:
            results (list[list | tuple]): Testing results of the dataset.
            metric (str | list[str]): Metrics to be evaluated. Options are
                'bbox', 'segm', 'proposal', 'proposal_fast'.
            logger (logging.Logger | str | None): Logger used for printing
                related information during evaluation. Default: None.
            jsonfile_prefix (str | None): The prefix of json files. It includes
                the file path and the prefix of filename, e.g., "a/b/prefix".
                If not specified, a temp file will be created. Default: None.
            classwise (bool): Whether to evaluating the AP for each class.
            proposal_nums (Sequence[int]): Proposal number used for evaluating
                recalls, such as recall@100, recall@1000.
                Default: (100, 300, 1000).
            iou_thrs (Sequence[float], optional): IoU threshold used for
                evaluating recalls/mAPs. If set to a list, the average of all
                IoUs will also be computed. If not specified, [0.50, 0.55,
                0.60, 0.65, 0.70, 0.75, 0.80, 0.85, 0.90, 0.95] will be used.
                Default: None.
            metric_items (list[str] | str, optional): Metric items that will
                be returned. If not specified, ``['AR@100', 'AR@300',
                'AR@1000', 'AR_s@1000', 'AR_m@1000', 'AR_l@1000' ]`` will be
                used when ``metric=='proposal'``, ``['mAP', 'mAP_50', 'mAP_75',
                'mAP_s', 'mAP_m', 'mAP_l']`` will be used when
                ``metric=='bbox' or metric=='segm'``.

        Returns:
            dict[str, float]: COCO style evaluation metric.
        """

        metrics = metric if isinstance(metric, list) else [metric]
        allowed_metrics = ['bbox', 'segm', 'proposal', 'proposal_fast']
        for metric in metrics:
            if metric not in allowed_metrics:
                raise KeyError(f'metric {metric} is not supported')

        coco_gt = self.coco
        self.cat_ids = coco_gt.get_cat_ids(cat_names=self.CLASSES)

        result_files, tmp_dir = self.format_results(results, jsonfile_prefix)
        eval_results = self.evaluate_det_segm(results, result_files, coco_gt,
                                              metrics, logger, classwise,
                                              proposal_nums, iou_thrs,
                                              metric_items)

        if tmp_dir is not None:
            tmp_dir.cleanup()
        return eval_results





class ValidationError(RuntimeError):
    pass


class StructHasDefinedError(RuntimeError):
    pass


class StructNotFoundError(RuntimeError):
    pass


class DefineSyntaxError(SyntaxError):
    pass


class DefineTypeError(TypeError):
    pass



import json
import re


class DetectionParse:
    PARAM_NAME = "cdom"
    FIELD_MAPPING = {
        "bbox": "BBox",
        "int": "Int",
        "float": "Num",
        "str": "Str",
        "points": "Polygon",
        "list": "List",
        "media": "Image",
        "bool": "Bool",
        "category": f"Label[dom=${PARAM_NAME}]",
    }

    def __init__(self, data_info_file, ann_file, dataset_name=None):
        self._dataset_name = dataset_name
        self._data_info_file = data_info_file
        self._annotation_file = ann_file
        self._data_info = self._read_json(data_info_file)
        self._annotation_info = self._read_json(ann_file)
        self._meta_info = self._parse_meta_info()
        self._class_domain_info, self._class_raw_info = self._parse_class_domain()
        self._dsdl_version_info = {"$dsdl-version": "0.5.0"}
        self._struct_defs, self._samples = self._parse_ann_info()

    def _parse_meta_info(self):
        meta_info = {k: v for k, v in self._data_info.items() if k not in ("tasks", "statistics")}
        meta_info["sub_dataset_name"] = self._annotation_info["sub_dataset_name"]
        if not self._dataset_name:
            self._dataset_name = meta_info["dataset_name"]
        return meta_info

    def _parse_class_domain(self):
        class_raw_info = {}
        class_domain_info = {"$def": "class_domain", "$name": self.clean(f"{self._dataset_name}ClassDom")}
        classes = []
        for task in self._data_info["tasks"]:
            task_name, task_type, task_source, catalog = task["name"], task["type"], task["source"], task[
                "catalog"]
            task_id = f"{task_name}-{task_type}-{task_source}"
            class_raw_info[task_id] = {_["category_id"]: _["category_name"] for _ in catalog}
            classes.extend(class_raw_info[task_id].values())
        class_domain_info["classes"] = sorted(list(set(classes)))
        return class_domain_info, class_raw_info

    def _parse_ann_info(self):
        samples = self._annotation_info["samples"]
        ann_info = []
        object_struct = {
            "$name": "LocalObjectEntry",
            "$def": "struct",
            "$params": [self.PARAM_NAME],
            "$fields": {
            }
        }
        sample_struct = {
            "$name": "ObjectDetectionSample",
            "$def": "struct",
            "$params": [self.PARAM_NAME],
            "$fields": {
                "media": self.FIELD_MAPPING["media"],
                "annotations": f"List[etype={object_struct['$name']}[{self.PARAM_NAME}=${self.PARAM_NAME}]]"
            }
        }
        data_info = {"sample-type": f"{sample_struct['$name']}[{self.PARAM_NAME}={self._class_domain_info['$name']}]"}
        optional_fields = set()
        annotations_optional_flag = False
        empty_annotations_flag = True
        for sample in samples:
            ground_truth_dict = {}
            sample_item = {"media": sample["media"]["path"]}
            for gt_item in sample.get("ground_truth", []):
                ann_id = gt_item.get("ref_ann_id") or gt_item["ann_id"]
                ground_truth_dict.setdefault(ann_id, {})

                if "bbox" in gt_item:
                    object_struct["$fields"]["bbox"] = self.FIELD_MAPPING["bbox"]
                    ground_truth_dict[ann_id]["bbox"] = gt_item["bbox"]

                if "points" in gt_item:
                    object_struct["$fields"]["points"] = self.FIELD_MAPPING["points"]
                    ground_truth_dict[ann_id]["points"] = gt_item["points"]

                task_id = f"{gt_item['name']}-{gt_item['type']}-{gt_item['source']}"
                _category_name = self._class_raw_info[task_id][gt_item["categories"][0]["category_id"]]
                ground_truth_dict[ann_id]["category"] = self._class_domain_info["classes"].index(_category_name) + 1
                object_struct["$fields"]['category'] = self.FIELD_MAPPING['category']
                ground_truth_dict[ann_id].update(gt_item["attributes"])
                _attribute_field_dict = {k: self.FIELD_MAPPING[v.__class__.__name__] for k, v in
                                         gt_item.get("attributes", {}).items() if
                                         v.__class__.__name__ in self.FIELD_MAPPING}
                object_struct["$fields"].update(_attribute_field_dict)
                optional_fields.update(_attribute_field_dict.keys())
            ground_truths = list(ground_truth_dict.values())
            if ground_truths:
                empty_annotations_flag = False
                sample_item["annotations"] = ground_truths
            else:
                annotations_optional_flag = True
            ann_info.append(sample_item)
        data_info["samples"] = ann_info
        if optional_fields:
            object_struct["$optional"] = list(optional_fields)
        if empty_annotations_flag:
            sample_struct["$fields"].pop("annotations")
            struct_defs = [sample_struct]
        else:
            if annotations_optional_flag:
                sample_struct["$optional"] = ["annotations"]
            struct_defs = [object_struct, sample_struct]
        return struct_defs, data_info

    @staticmethod
    def _read_json(file):
        with open(file, "r") as f:
            json_info = json.load(f)
        return json_info

    @staticmethod
    def clean(varStr):
        return re.sub('\W|^(?=\d)', '', varStr)

    @property
    def struct_defs(self):
        """
        dsdl的struct定义字典组成的列表
        """
        return self._struct_defs

    @property
    def samples(self):
        """
        dsdl的data字段下的内容，为一个字典，有sample-type域samples字段，samples包含了样本列表
        """
        return self._samples

    @property
    def sample_info(self):
        """
        tunas v0.3的annotation的json文件的原始内容
        """
        return self._annotation_info

    @property
    def class_domain_info(self):
        """
        dsdl的class_domain的定义字典
        """
        return self._class_domain_info

    @property
    def meta_info(self):
        """
        dsdl的元信息内容
        """
        return self._meta_info

    @property
    def data_info_file(self):
        """
        返回tunas v0.3 dataset_info.json的文件路径
        """
        return self._data_info_file

    @property
    def annotation_file(self):
        """
        返回tunas v0.3 的标注文件json文件的路径
        """
        return self._annotation_file

    @property
    def dsdl_version(self):
        """
        dsdl的版本信息字典
        """
        return self._dsdl_version_info
